# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Markdown to HTML conversion."""

import argparse
import re
import sys
import tempfile

import markdown

__all__ = ("convert_markdown",)


HTML_HEADER = """\
<!DOCTYPE html>
<html xmlns="http://www.w3.org/1999/xhtml" lang="" xml:lang="">
<head>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />
  <title>Preview</title>
  <style>
    html {
      line-height: 1.3;
      font-family: IBM Plex Sans, Georgia, serif;
      font-size: 10pt;
    }
    h1, h2 {
      margin-top: 1cm;
      border-bottom: 1pt solid #DDDDDD;
    }
    h3, h4, p {
      margin-top: 0.5cm;
      padding-left: 1.0cm;
    }
    ul {
      padding-left: 2.0cm;
    }
    ul ul {
      padding-left: 1.0cm;
    }
    li {
      margin-top: 2pt;
      margin-bottom: 2pt;
    }
    code {
      font-family: IBM Plex Mono;
      background-color: rgba(175, 184, 193, 0.2);
      padding: 2pt 5pt 2pt 5pt;
      border-radius: 6pt;
    }
    table {
      margin-left: auto;
      margin-right: auto;
      text-align: center;
      border-collapse: collapse;
    }
    td, th {
      border: 1pt solid #BBBBBB;
      padding: 4pt;
      font-size: 10pt;
    }
    @page {
        size: A4;
        margin: 1.5cm 1.5cm 1.5cm 1.5cm;
    }
  </style>
</head>
<body>
"""


HTML_FOOTER = "</body>"


MACRO_TEXT = r"""\
\bvec:\vec{\mathbf{#1}}
\normvec:\hat{\mathbf{#1}}
\d:\operatorname{d}\!{#1}
\ihat:\hat{\mathbf{i}}
\jhat:\hat{\mathbf{j}}
\khat:\hat{\mathbf{k}}
"""


def convert_markdown(text_md: str) -> str:
    """Convert Markdown to HTML with KaTeX support.

    Parameters
    ----------
    text_md
        The markdown source text

    Returns
    -------
    html
        The HTML conversion.
    """
    # Convert conventional LaTeX equation syntax to make it compatible with markdown_katex
    text_md = re.sub(r"\B\$(\S|\S[^\n\r]*?\S)\$\B", r"$`\1`$", text_md)

    # Write macros to temporary file for KaTeX.
    with tempfile.NamedTemporaryFile(suffix=".tex") as f:
        f.write(MACRO_TEXT.encode("ascii"))
        f.flush()
        fn_macro = f.name

        md_ctx = markdown.Markdown(
            extensions=[
                "fenced_code",
                "markdown_katex",
                "tables",
            ],
            extension_configs={
                "markdown_katex": {"insert_fonts_css": True, "macro-file": fn_macro}
            },
        )

        # Convert to HTML
        text_html = HTML_HEADER + md_ctx.convert(text_md) + HTML_FOOTER

    return text_html


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-convert-markdown", description="Convert Markdown to HTML"
    )
    parser.add_argument("markdown", help="A Markdown file with extension `.md`")
    parser.add_argument("html", help="A HTML output filename")
    return parser.parse_args()


def main() -> int:
    """Main program."""
    args = parse_args()
    if not args.markdown.endswith(".md"):
        raise ValueError("The markdown file must end with the .md extension.")
    with open(args.markdown) as fm, open(args.html, "w") as fh:
        fh.write(convert_markdown(fm.read()))
    return 0


if __name__ == "__main__":
    sys.exit(main())



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Create ZIP with all files listed in a MANIFEST.txt file."""

import argparse
import datetime
import sys
import tempfile
import zipfile

from path import Path

from stepup.core.hash import compute_file_digest

from .check_manifest import iter_manifest

__all__ = ("zip_manifest",)


TIMESTAMP = datetime.datetime(1980, 1, 1).timestamp()


def main() -> int:
    """Main program."""
    args = parse_args()
    return zip_manifest(args.manifest_txt, args.output_zip)


def parse_args() -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-zip-manifest", description="Create a reproducible ZIP file."
    )
    parser.add_argument(
        "manifest_txt",
        help="The MANIFEST.txt with all files to be zipped. "
        "The digests of the files will be checked before archiving. "
        "The manifest file will be included in the ZIP.",
    )
    parser.add_argument("output_zip", help="Destination zip file.")
    return parser.parse_args()


def zip_manifest(path_manifest: str, path_zip: str, check: bool = True) -> int:
    """Create a reproducible zip file.

    Parameters
    ----------
    path_manifest
        The MANIFEST.txt file
    path_zip
        The ZIP file to be created.
    check
        When True, the file will be validated with size and hash
        before adding it to the ZIP file.
        In case of a mismatch, the partial ZIP is removed.
    """
    if not path_zip.endswith(".zip"):
        print(f"Destination must have a `.zip` extension. Got {path_zip}", file=sys.stderr)
        return 2
    path_zip = Path(path_zip)
    if not path_manifest.endswith(".txt"):
        print(
            f"Manifest file must have a `.sha256` extension. Got {path_manifest}", file=sys.stderr
        )
        return 2
    path_manifest = Path(path_manifest)

    # Create new one.
    root = path_manifest.parent
    nskip = 0 if root == "" else len(root) + 1
    result = 0
    with tempfile.TemporaryDirectory("reprep-zip-manifest") as tmpdir:
        tmpdir = Path(tmpdir)
        path_zip_tmp = tmpdir / "out.zip"
        with zipfile.ZipFile(path_zip_tmp, "w") as fz:
            for size, digest, src in iter_manifest(path_manifest):
                # Copy the file to a temp dir.
                # This saves bandwith in case of remote datasets and allows
                # fixing the timestamp before compression.
                dst = tmpdir / "todo"
                src.copy(dst)
                dst.utime((TIMESTAMP, TIMESTAMP))
                # Check if needed
                if check:
                    actual_size = Path(dst).getsize()
                    if size != actual_size:
                        print(
                            f"Size {actual_size} of {src} differs from expected size {size}",
                            file=sys.stderr,
                        )
                        result = 2
                        break
                    actual_digest = compute_file_digest(src)
                    if digest != actual_digest:
                        print(f"Digest mismatch for file: {src}", file=sys.stderr)
                        result = 3
                        break
                # Compress
                fz.write(dst, src[nskip:], zipfile.ZIP_DEFLATED)
            dst = tmpdir / "todo"
            path_manifest.copy(dst)
            dst.utime((TIMESTAMP, TIMESTAMP))
            fz.write(dst, path_manifest[nskip:], zipfile.ZIP_DEFLATED)

        if result == 0:
            path_zip_tmp.move(path_zip)
    return result


if __name__ == "__main__":
    sys.exit(main())



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Checking of MANIFEST files.

There are two kinds of MANIFEST files:

- ``MANIFEST.in`` contains rules to select or ignore files to be included into an archive.
- ``MANIFEST.txt`` is the result of a processing a ``MANIFEST.in`` file with this script.
  It contains every individual file, together with its size in bytes and its blake2b hexdigest.
"""

import argparse
import sys
from collections.abc import Iterator

from path import Path

from stepup.core.hash import compute_file_digest

__all__ = ("main", "iter_manifest", "check_manifest")


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if not args.manifest_txt.endswith(".txt"):
        raise ValueError("The manifest output file must end with .txt")
    check_manifest(args.manifest_txt)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="reprep-check-manifest", description="Check a MANIFEST.txt."
    )
    parser.add_argument("manifest_txt", help="A MANIFEST.txt file generated with RepRep")
    args = parser.parse_args(argv)
    return args


def iter_manifest(path_manifest: str) -> Iterator[tuple[int, bytes, str]]:
    root = path_manifest.parent.normpath()
    with open(path_manifest) as fh:
        for iline, line in enumerate(fh):
            if len(line) < 146:
                raise ValueError(f"Line {iline} too short in {path_manifest}")
            size = int(line[:15])
            digest = bytes.fromhex(line[16:144])
            path = line[145:].strip()
            yield size, digest, root / path


def check_manifest(path_manifest: str):
    path_manifest = Path(path_manifest)
    for size, digest, path in iter_manifest(path_manifest):
        actual_size = Path(path).getsize()
        if size != actual_size:
            raise ValueError(f"File size should be {size} but got {actual_size}: {path}")
        actual_digest = compute_file_digest(path)
        if digest != actual_digest:
            raise ValueError(f"File digest mismatch: {path}")


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# Copyright (C) 2024 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Unit tests for manifest files."""

import contextlib

from stepup.reprep.check_manifest import main as check_main
from stepup.reprep.make_manifest import main as make_main

BASIC_MANIFEST = """\
              4 b25ac67550816e84f2a5689115534c8b1d72081f14f6bc4395e4af41c1dfa83163d5433ae05e4cba54e\
04364d38f9084ec9c3e41d0fbe45da3b738939b1aec11 a.txt
              4 10901f8827ee64f82d48afd9fc4a9bf9befaea8cf0b00a4bf16b1533cae88c04f7b467b4c59a690462e\
898e75a0def6c0168b3aef7d01600944f811b4b75da35 b.txt
"""


def test_basic(tmpdir):
    with contextlib.chdir(tmpdir):
        with open("a.txt", "w") as fh:
            print("aaa", file=fh)
        with open("b.txt", "w") as fh:
            print("bbb", file=fh)
        with open("MANIFEST.in", "w") as fh:
            print("include *.txt", file=fh)
        make_main(["-i", "MANIFEST.in"])
        with open("MANIFEST.txt") as fh:
            assert fh.read() == BASIC_MANIFEST
        make_main(["-o", "MANIFEST2.txt", "a.txt", "b.txt"])
        with open("MANIFEST2.txt") as fh:
            assert fh.read() == BASIC_MANIFEST
        check_main(["MANIFEST.txt"])
